/**
 * TCS Dependency Map Renderer
 *
 * Copyright (c) 2005 TopCoder, Inc. All rights Reserved.
 */
package com.topcoder.rules.dependencymap.view;

import java.awt.Point;
import java.awt.Insets;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.Color;

import java.awt.Rectangle;
import java.awt.geom.GeneralPath;
import java.awt.geom.AffineTransform;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import com.topcoder.util.config.ConfigManager;


/**
 * This class is an implementation of the LayoutManager that will layout the dependency diagram into a fixed
 * sized grid (where all columns are fixed to a single width and all rows are fixed to a single height).
 *
 * @author Pops
 * @author TCSDeveloper
 */
public class FixedGridLayoutManager implements LayoutManager {
    
    
    
    /**
     * This variable represents the extra padding that should appear between columns.
     * This variable is only used *between* columns and will not provide margins
     * around the outer columns.  This variable is set in the constructor, is
     * mutatable by the setter, will be >= 0 and will be used by the layout mechanisms.
     */
    private int colPadding = 10;

    /**
     * This variable represents the extra padding that should appear between rows.
     * This variable is only used *between* rows and will not provide margins
     * around the outer rows.  This variable is set in the constructor, is
     * mutatable by the setter, will be >= 0 and will be used by the layout mechanisms.
     */
    private int rowPadding = 10;

    /**
     * This variable represents the extra padding that should appear around
     * text (top/bottom/left/right).  This variable is set in the constructor, is
     * mutatable by the setter, will be >= 0 and will be used by the layout mechanisms.
     */
    private int textMargin = 10;

    /**
     * This variable represents the text font that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Font textFont = null;

    /**
     * This variable represents the text color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color textColor = new Color(0x000000);

    /**
     * This variable represents the shape line color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color shapeLineColor = new Color(0x000000);

    /**
     * This variable represents the shape fill color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color shapeFillColor = new Color(0xffffff);

    /**
     * This variable represents the arrow line color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color arrowLineColor = new Color(0x000000);

    /**
     * This variable represents the arrow fill color that should be used.
     * This variable is set in the constructor, is
     * mutatable by the setter, will never be null, and will be used
     * by the layout mechanisms.
     */
    private Color arrowFillColor = new Color(0xffffff);
        
    /**
     * Constructs the fixed grid layout manager using the classname as
     * the default namespace.  This constructor simply calls 
     * this("com.topcoder.rules.dependencymap.view.FixedGridLayoutManager")
     */
    public FixedGridLayoutManager() { 
        this("com.topcoder.rules.dependencymap.view.FixedGridLayoutManager");
        // note: "this(getClass().getName())" does not work.
    }
    
    /**
     * Constructs the FixedGridLayoutManager from the given namespace.
     * This method simply saves the namespace to the like named variable.
     * 
     * 
     * 
     * @param namespace The namespace to use
     * @throws NullPointerException if namespace is null
     * @throws IllegalArgumentException if namespace (trim'd) is empty
     */
    public FixedGridLayoutManager(String namespace) {
        
        if (namespace == null) {
            throw new NullPointerException("namespace should not be null.");
        }
        if (namespace.trim().length() == 0) {
            throw new IllegalArgumentException("namespace should not be empty.");
        }
        try {
        
            ConfigManager manager = ConfigManager.getInstance();
            
            colPadding = getInt(manager, namespace, "colPadding", 10);
            rowPadding = getInt(manager, namespace, "rowPadding", 10);
            textMargin = getInt(manager, namespace, "textMargin", 10);
            
            // retreave colors
            textColor = getColor(manager, namespace, "textColor", 0x000000);
            shapeLineColor = getColor(manager, namespace, "shapeLineColor", 0x000000);
            shapeFillColor = getColor(manager, namespace, "shapeFillColor", 0xffffff);
            arrowLineColor = getColor(manager, namespace, "arrowLineColor", 0x000000);
            arrowFillColor = getColor(manager, namespace, "arrowFillColor", 0xffffff);

            // retreave font                        
            int fontSize = getInt(manager, namespace, "textFontSize", 12);
            if (fontSize < 8) {
                fontSize = 12;
            }
            
            String textFontStyle = manager.getString(namespace, "textFontStyle");
            String textFontName = manager.getString(namespace, "textFontName");
            
            textFont = getFont(textFontStyle, textFontName, fontSize);
        } catch (Exception e) {
            textFont = new Font("SansSerif", Font.PLAIN, 12);            
        }
        
        
    }
    
    /**
     * Get the int property.
     * 
     * 
     * @param manager the ConfigManager instance
     * @param name the property name
     * @param namespace The namespace to use
     * @param defaultValue The default value     
     * @return the int value.
     */
    private int getInt(ConfigManager manager, String namespace, String name, int defaultValue) {
        try {
            String stringValue = manager.getString(namespace, name);
            int value = Integer.parseInt(stringValue);                        
            if (value < 0) {
                return defaultValue;
            }
            return value;
            
        } catch (Exception e) {
            return defaultValue;
        }            
        
    }
    
    /**
     * Get the Color property.
     * 
     * 
     * @param manager the ConfigManager instance
     * @param name the property name
     * @param namespace The namespace to use
     * @return the Color value.
     */
    private Color getColor(ConfigManager manager, String namespace, String name, int defaultValue) {
        try {
            String stringValue = manager.getString(namespace, name);
            int value = Integer.parseInt(stringValue);                        
            if (value < 0) {
                value = defaultValue;
            }
            return new Color(value);
            
        } catch (Exception e) {
            return new Color(defaultValue);
        }            
        
    }

    /**
     * Get the font.
     * 
     * 
     * @param name the font name
     * @param style the font style
     * @param size the font size
     * @return the font
     */
     private Font getFont(String name, String style, int size) {
        
        if (size < 8) {
            return new Font("SansSerif", Font.PLAIN, 12);
        }
        
        int s = 0;
        
        if (style == "Plain") {
            s = Font.PLAIN;
        } else if (style == "Bold") {
            s = Font.BOLD;
        } else if (style == "Italic") {
            s = Font.ITALIC;
        } else if (style == "Bold/Italic") {
            s = Font.PLAIN | Font.ITALIC;
        } else {
            return new Font("SansSerif", Font.PLAIN, 12);
        }
        
        return new Font(name, s, size);         
        
    }
        
    
    
    /**
     * Calculate the widths of each of the unseen dependencies with is referenced by a root dependency.
     *
     * @param dependency the root dependency
     * @return a mapping from dependency to widths (Integer)
     */
    private Map getWidths(Dependency dependency) {
        // Mapping from dependants to widths (number of cells)
        Map widths = new HashMap();
        // Visitation count for dependants (Integer)
        Map seen = new HashMap();
        // BFS queue
        List queue = new ArrayList();

        // Initialize the queue and visitation count by placing the root dependency.
        queue.add(dependency);
        seen.put(dependency, new Integer(1));

        // Iterate through the queue.
        for (int i = 0; i < queue.size(); ++i) {
            dependency = (Dependency) queue.get(i);

            for (Iterator itr = dependency.getDependants().iterator(); itr.hasNext();) {
                Dependency dependant = (Dependency) itr.next();

                // If the dependency is seen for the first time, we need to calculate its
                // width, otherwise it's a seen dependency and has constant width of 1.
                if (!seen.containsKey(dependant)) {
                    queue.add(dependant);
                    seen.put(dependant, new Integer(1));
                } else {
                    seen.put(dependant, new Integer(((Integer) seen.get(dependant)).intValue() + 1));
                }
            }
        }

        // Calculate the width in a backward order.
        for (int i = queue.size() - 1; i >= 0; --i) {
            dependency = (Dependency) queue.get(i);

            int width = 0;
            for (Iterator itr = dependency.getDependants().iterator(); itr.hasNext();) {
                Dependency dependant = (Dependency) itr.next();

                // If the dependency is an unseen dependency, its width should already been calculated.
                // Otherwise its width is constantly 1.
                if (seen.get(dependant).equals(new Integer(1))) {
                    width += ((Integer) widths.get(dependant)).intValue();
                } else {
                    ++width;
                    seen.put(dependant, new Integer(((Integer) seen.get(dependant)).intValue() - 1));
                }
            }

            // If width is even make it odd.
            if (width % 2 == 0) {
                ++width;
            }

            // Store the width into the mapping.
            widths.put(dependency, new Integer(width));
        }

        return widths;
    }

    /**
     * Paint a dependency.
     *
     * @param dependency the dependency to paint
     * @param shape the shape for outbound
     * @param width the width of a cell
     * @param height the height of a cell
     * @param arrowHeight the height of an arrow
     * @param x the x coord position
     * @param y the y coord position
     * @param graphics the graphics to paint onto
     */
    private void paintDependency(Dependency dependency, GeneralPath shape, int width, int height,
                                 int arrowHeight, int x, int y, RendererGraphics graphics) {
        // Calculate pixel offset.
        x *= width + this.colPadding;
        y *= height + arrowHeight + this.rowPadding;

        // Scale it to cell size and translate it to position.
        Rectangle rect = shape.getBounds();
        shape.transform(AffineTransform.getScaleInstance(((double) width) / rect.width,
                                                         ((double) height) / rect.height));
        shape.transform(AffineTransform.getTranslateInstance(x, y));

        // Draw the shape.
        graphics.fill(shape);        
        graphics.draw(shape);
        

        // Retrieve font metrics and estimate text width.
        FontMetrics metrics = graphics.getFontMetrics(graphics.getFont());
        int myTextWidth = metrics.stringWidth(dependency.getTitle());

        // Draw the text to the center of the cell.
        graphics.drawString(dependency.getTitle(), x + this.textMargin +
                            (width - this.textMargin - this.textMargin - myTextWidth) / 2,
                                                   y + this.textMargin + 1 + metrics.getAscent());
    }

    /**
     * Paint an arrow.
     *
     * @param shape the shape for the arrow
     * @param x1 the x coord of the first dependency
     * @param x2 the x coord of the second dependency
     * @param y the y coord of the first dependency
     * @param width the width of a cell
     * @param height the height of a cell
     * @param graphics the graphics to paint onto
     */
    private void paintArrow(GeneralPath shape, int x1, int x2, int y,
                            int width, int height, RendererGraphics graphics) {
        Rectangle rect = shape.getBounds();
        // Calculate arrow position.
        float x = (float) ((x1 + x2) / 2.0);
        x *= width + this.colPadding;
        x += (width - rect.width) / 2;
        y *= height + rect.height + this.rowPadding;
        y += height + this.rowPadding / 2;

        // Rotate it to point the dependant and translate it to position.
        shape.transform(AffineTransform.getRotateInstance(
                Math.PI / 2 - Math.atan2(height + rect.height + this.rowPadding,
                                         (x1 - x2) * (width + this.colPadding)),
                rect.width / 2.0, rect.height / 2.0));
        shape.transform(AffineTransform.getTranslateInstance(x, y));

        // Draw the arrow.
        graphics.draw(shape);
        
    }

    /**
     * Paint the dependency hiearchy.
     *
     * @param dependency the root dependency
     * @param widths the mapping of the width of the dependencies
     * @param shapeFactory the shape factory to produce the shapes
     * @param graphics the graphics to paint onto
     */
    private void paint(Dependency dependency, Map widths, ShapeFactory shapeFactory, RendererGraphics graphics) {
        
        graphics.setColor(shapeFillColor);
        graphics.setFont(textFont);
        
        // Calculate the maximum text size.
        FontMetrics metrics = graphics.getFontMetrics(graphics.getFont());
        int textHeight = metrics.getHeight();
        int textWidth = 0;
        for (Iterator itr = widths.keySet().iterator(); itr.hasNext();) {
            int myTextWidth = metrics.stringWidth(((Dependency) itr.next()).getTitle());

            if (myTextWidth > textWidth) {
                textWidth = myTextWidth;
            }
        }
        int height = textHeight + this.textMargin + this.textMargin + 2;
        int width = textWidth + this.textMargin + this.textMargin + 2;
        int arrowHeight = shapeFactory.getShape(ShapeFactory.RELATIONSHIP_ARROW).getBounds().height;

        // Visitation set
        Set seen = new HashSet();
        // BFS queue
        List queue = new ArrayList();
        // Position of the dependency to render
        List positions = new ArrayList();
        // The position of the parent dependency
        List parentPos = new ArrayList();

        // Initialize by inserting the root dependency.
        seen.add(dependency);
        queue.add(dependency);
        positions.add(new Point(0, 0));
        parentPos.add(null);

        // Render through the queue.
        for (int i = 0; i < queue.size(); ++i) {
            dependency = (Dependency) queue.get(i);
            Point position = (Point) positions.get(i);

            // Calculate the dependants width to decide whether we need a padding cell.
            int dependantsWidth = 0;
            for (Iterator itr = dependency.getDependants().iterator(); itr.hasNext();) {
                Dependency dependant = (Dependency) itr.next();

                if (!seen.contains(dependant)) {
                    dependantsWidth += ((Integer) widths.get(dependant)).intValue();
                } else {
                    ++dependantsWidth;
                }
            }
            boolean padding = dependantsWidth % 2 == 0;

            // If parent dependency exists we make the relationship arrow.
            if (parentPos.get(i) != null) {
                int parent = ((Integer) parentPos.get(i)).intValue();
                paintArrow(shapeFactory.getShape(ShapeFactory.RELATIONSHIP_ARROW),
                           parent, position.x + dependantsWidth / 2, position.y - 1, width, height, graphics);
            }

            // Obtain shape used for this unseen shape.
            GeneralPath shape = null;
            if (dependency.getShapeOverride() != -1) {
                shape = shapeFactory.getShape(dependency.getShapeOverride());
            }
            if (shape == null) {
                shape = shapeFactory.getShape(ShapeFactory.UNSEEN_ELEMENT);
            }
            paintDependency(dependency, shape, width, height,
                            arrowHeight, position.x + dependantsWidth / 2, position.y, graphics);

            // Accumulate the width to find the position for each dependant.
            int cumulativeWidth = 0;
            for (Iterator itr = dependency.getDependants().iterator(); itr.hasNext();) {
                Dependency dependant = (Dependency) itr.next();

                // Check to see if we make the padding on this dependant.
                if (padding && cumulativeWidth * 2 >= dependantsWidth) {
                    padding = false;
                    ++cumulativeWidth;
                }

                if (!seen.contains(dependant)) {

                    // Push the unseen dependant into the queue.
                    seen.add(dependant);
                    queue.add(dependant);
                    positions.add(new Point(position.x + cumulativeWidth, position.y + 1));
                    parentPos.add(new Integer(position.x + dependantsWidth / 2));

                    cumulativeWidth += ((Integer) widths.get(dependant)).intValue();
                } else {
                    // Draw arrow for the seen dependant.
                    paintArrow(shapeFactory.getShape(ShapeFactory.RELATIONSHIP_ARROW),
                               position.x + dependantsWidth / 2, position.x + cumulativeWidth, position.y,
                               width, height, graphics);

                    // Obtain shape used for this seen shape.
                    shape = null;
                    if (dependant.getShapeOverride() != -1) {
                        shape = shapeFactory.getShape(dependant.getShapeOverride());
                    }
                    if (shape == null) {
                        shape = shapeFactory.getShape(ShapeFactory.SEEN_ELEMENT);
                    }

                    // Draw the dependant.
                    paintDependency(dependant, shape, width, height,
                                    arrowHeight, position.x + cumulativeWidth, position.y + 1, graphics);

                    ++cumulativeWidth;
                }
            }
        }
    }

    /**
     * Lays out and prints the dependency diagram. Since the algorithm to do this is beyond Poseidon's
     * capabilities - please review the component specifications for details.
     *
     * @param dependency The dependency to layout
     * @param shapeFactory The shape factory to use
     * @param graphics The graphics to print to
     * @throws NullPointerException if either dependency, shapeFactory or g2d is null
      * @throws RendererException if dependency contains non-Dependency dependencies (tongue twister eh)
     */
    public void layout(Dependency dependency, ShapeFactory shapeFactory, DiagramRenderer renderer) throws RendererException {
        if (dependency == null) {
            throw new NullPointerException("dependency should not be null.");
        }
        if (shapeFactory == null) {
            throw new NullPointerException("shapeFactory should not be null.");
        }
        if (renderer == null) {
            throw new NullPointerException("renderer should not be null.");
        }
        
        RendererGraphics graphics = null;
               
        try {
            Map widths = getWidths(dependency);
            
            // first pass
            graphics = renderer.getGraphics(dependency, 1, 1);
            
            paint(dependency, widths, shapeFactory, graphics);
            
            graphics.dispose();
           
            // second pass
            graphics = renderer.getGraphics(dependency, 1, 1);
            
            paint(dependency, widths, shapeFactory, graphics);
            
            graphics.complete();
            
                   
        } catch (RendererException re) {
            throw re;
        } catch (Exception ex) {
            throw new RendererException("error occurs while layout", ex);
        } finally {
            graphics.dispose();            
        }
    }
    
    /**
     * This is the getter method for the colPadding variable.
     * 
     * 
     * 
     * @return the colPadding variable.
     */
    public int getColPadding() {        
        return colPadding;
    } 

    /**
     * This method is simply the setter method for the colPadding variable.
     * 
     * 
     * 
     * @param colPadding The column padding
     * @throws IllegalArgumentException if colPadding is < 0
     */
    public void setColPadding(int colPadding) {        
        if (colPadding < 0) {
            throw new IllegalArgumentException("colPadding should not be less than 0.");
        }
        this.colPadding = colPadding;
    } 

    /**
     * This is the getter method for the rowPadding variable.
     * 
     * 
     * 
     * @return the rowPadding variable
     */
    public int getRowPadding() {        
        return rowPadding;
    } 

    /**
     * This method is simply the setter method for the rowPadding variable.
     * 
     * 
     * 
     * @param rowPadding The row padding
     * @throws IllegalArgumentException if rowPadding is < 0
     */
    public void setRowPadding(int rowPadding) {        
        if (rowPadding < 0) {
            throw new IllegalArgumentException("rowPadding should not be less than 0.");
        }
        this.rowPadding = rowPadding;
    } 

    /**
     * This is the getter method for the textMargin variable.
     * 
     * 
     * 
     * @return the textMargin variable. 
     */
    public int getTextMargin() {        
        return textMargin;
    } 

    /**
     * This method is simply the setter method for the textMargin variable.
     * 
     * 
     * 
     * @param textMargin The margin around text
     * @throws IllegalArgumentException if textMargin is < 0
     */
    public void setTextMargin(int textMargin) {        
        if (textMargin < 0) {
            throw new IllegalArgumentException("textFont should not be less than 0.");
        }
        this.textMargin = textMargin;
    } 

    /**
     * This is the getter method for the textFont variable.
     * 
     * 
     * 
     * @return the textFont variable
     */
    public Font getTextFont() {        
        return textFont;
    } 

    /**
     * This method is simply the setter method for the textFont variable.
     * 
     * 
     * 
     * @param textFont The font for text
     * @throws NullPointerException if textFont is null
     */
    public void setTextFont(Font textFont) {        
        if (textFont == null) {
            throw new NullPointerException("textFont should not be null.");
        }
        this.textFont = textFont;
    } 

    /**
     * This is the getter method for the textColor variable.
     * 
     * 
     * 
     * @return the textColor variable.
     */
    public Color getTextColor() {        
        return textColor;
    } 

    /**
     * This method is simply the setter method for the textColor variable.
     * 
     * 
     * 
     * @param textColor The text color
     * @throws NullPointerException if textColor is null
     */
    public void setTextColor(Color textColor) {        
        if (textColor == null) {
            throw new NullPointerException("textColor should not be null.");
        }
        this.textColor = textColor;
    } 

    /**
     * This is the getter method for the shapeLineColor variable.
     * 
     * 
     * 
     * @return the shapeLineColor variable
     */
    public Color getShapeLineColor() {        
        return shapeLineColor;
    } 

    /**
     * This method is simply the setter method for the shapeLineColor variable.
     * 
     * 
     * 
     * @param shapeLineColor The shape's line color
     * @throws NullPointerException if shapeLineColor is null
     */
    public void setShapeLineColor(Color shapeLineColor) {        
        if (shapeLineColor == null) {
            throw new NullPointerException("shapeLineColor should not be null.");
        }
        this.shapeLineColor = shapeLineColor;
    } 

    /**
     * This is the getter method for the shapeFillColor variable.
     * 
     * 
     * 
     * @return The shape's fill color
     */
    public Color getShapeFillColor() {        
        return shapeFillColor;
    } 

    /**
     * This method is simply the setter method for the shapeFillColor variable.
     * 
     * 
     * 
     * @param shapeFillColor The shape's fill color
     * @throws NullPointerException if shapeFillColor is null
     */
    public void setShapeFillColor(Color shapeFillColor) {        
        if (shapeFillColor == null) {
            throw new NullPointerException("shapeFillColor should not be null.");
        }
        this.shapeFillColor = shapeFillColor;
    } 

    /**
     * This is the getter method for the arrowLineColor variable.
     * 
     * 
     * 
     * @return the arrowLineColor variabl
     */
    public Color getArrowLineColor() {        
        return arrowLineColor;
    } 

    /**
     * This method is simply the setter method for the arrowLineColor variable.
     * 
     * 
     * 
     * @param arrowLineColor The arrow line color
     * @throws NullPointerException if arrowLineColor is null
     */
    public void setArrowLineColor(Color arrowLineColor) {        
        if (arrowLineColor == null) {
            throw new NullPointerException("arrowLineColor should not be null.");
        }
        this.arrowLineColor = arrowLineColor;
    } 

    /**
     * This is the getter method for the arrowFillColor variable.
     * 
     * 
     * 
     * @return the arrowFillColor variable
     */
    public Color getArrowFillColor() {        
        return arrowFillColor;
    } 

    /**
     * This method is simply the setter method for the arrowFillColor variable.
     * 
     * 
     * 
     * @param arrowFillColor The fill color for arrows
     * @throws NullPointerException if arrowFillColor is null
     */
    public void setArrowFillColor(Color arrowFillColor) {  
        if (arrowFillColor == null) {
            throw new NullPointerException("arrowFillColor should not be null.");
        }
        this.arrowFillColor = arrowFillColor;
    }    
        
        
}
